@@ -17,9 +17,9 @@
 use crate::{
     impl_sw_curve_serializer,
     impl_sw_from_random_bytes,
-    traits::{AffineCurve, ProjectiveCurve, SWModelParameters as Parameters},
+    traits::{AffineCurve, Group, ProjectiveCurve, SWModelParameters as Parameters},
 };
-use snarkvm_fields::{Field, One, PrimeField, SquareRootField, Zero};
+use snarkvm_fields::{impl_additive_ops_from_ref, Field, One, PrimeField, SquareRootField, Zero};
 use snarkvm_utilities::{
     bititerator::BitIteratorBE,
     bytes::{FromBytes, ToBytes},
@@ -80,24 +80,13 @@ impl<P: Parameters> GroupAffine<P> {
         self.mul_bits(BitIteratorBE::new(P::COFACTOR))
     }
 
-    pub(crate) fn mul_bits<S: AsRef<[u64]>>(&self, bits: BitIteratorBE<S>) -> GroupProjective<P> {
-        let mut res = GroupProjective::zero();
-        for i in bits {
-            res.double_in_place();
-            if i {
-                res.add_assign_mixed(self)
-            }
-        }
-        res
-    }
-
     pub fn is_on_curve(&self) -> bool {
         if self.is_zero() {
             true
         } else {
             // Check that the point is on the curve
             let y2 = self.y.square();
-            let x3b = P::add_b(&((self.x.square() * &self.x) + &P::mul_by_a(&self.x)));
+            let x3b = P::add_b(&((self.x.square() * self.x) + P::mul_by_a(&self.x)));
             y2 == x3b
         }
     }
@@ -118,7 +107,6 @@ impl<P: Parameters> Zero for GroupAffine<P> {
 impl<P: Parameters> AffineCurve for GroupAffine<P> {
     type BaseField = P::BaseField;
     type Projective = GroupProjective<P>;
-    type ScalarField = P::ScalarField;
 
     impl_sw_from_random_bytes!();
 
@@ -134,7 +122,7 @@ impl<P: Parameters> AffineCurve for GroupAffine<P> {
     /// largest y-coordinate be selected.
     fn from_x_coordinate(x: Self::BaseField, greatest: bool) -> Option<Self> {
         // Compute x^3 + ax + b
-        let x3b = P::add_b(&((x.square() * &x) + &P::mul_by_a(&x)));
+        let x3b = P::add_b(&((x.square() * x) + P::mul_by_a(&x)));
 
         x3b.sqrt().map(|y| {
             let negy = -y;
@@ -153,14 +141,15 @@ impl<P: Parameters> AffineCurve for GroupAffine<P> {
         unimplemented!()
     }
 
-    fn add(self, _other: &Self) -> Self {
-        unimplemented!()
-    }
-
-    #[inline]
-    fn mul<S: Into<<Self::ScalarField as PrimeField>::BigInteger>>(&self, by: S) -> GroupProjective<P> {
-        let bits = BitIteratorBE::new(by.into());
-        self.mul_bits(bits)
+    fn mul_bits<S: AsRef<[u64]>>(&self, bits: BitIteratorBE<S>) -> GroupProjective<P> {
+        let mut res = GroupProjective::zero();
+        for i in bits {
+            res.double_in_place();
+            if i {
+                res.add_assign_mixed(self)
+            }
+        }
+        res
     }
 
     fn mul_by_cofactor_to_projective(&self) -> Self::Projective {
@@ -196,12 +185,30 @@ impl<P: Parameters> AffineCurve for GroupAffine<P> {
         } else {
             // Check that the point is on the curve
             let y2 = self.y.square();
-            let x3b = P::add_b(&((self.x.square() * &self.x) + &P::mul_by_a(&self.x)));
+            let x3b = P::add_b(&((self.x.square() * self.x) + P::mul_by_a(&self.x)));
             y2 == x3b
         }
     }
 }
 
+impl<P: Parameters> Group for GroupAffine<P> {
+    type ScalarField = P::ScalarField;
+
+    #[inline]
+    #[must_use]
+    fn double(&self) -> Self {
+        let mut tmp = *self;
+        tmp += self;
+        tmp
+    }
+
+    #[inline]
+    fn double_in_place(&mut self) {
+        let tmp = *self;
+        *self = tmp.double();
+    }
+}
+
 impl<P: Parameters> Neg for GroupAffine<P> {
     type Output = Self;
 
@@ -215,6 +222,56 @@ impl<P: Parameters> Neg for GroupAffine<P> {
     }
 }
 
+impl_additive_ops_from_ref!(GroupAffine, Parameters);
+
+impl<'a, P: Parameters> Add<&'a Self> for GroupAffine<P> {
+    type Output = Self;
+
+    fn add(self, other: &'a Self) -> Self {
+        let mut copy = self;
+        copy += other;
+        copy
+    }
+}
+
+impl<'a, P: Parameters> AddAssign<&'a Self> for GroupAffine<P> {
+    fn add_assign(&mut self, other: &'a Self) {
+        let mut projective = GroupProjective::from(*self);
+        projective.add_assign_mixed(other);
+        *self = projective.into();
+    }
+}
+
+impl<'a, P: Parameters> Sub<&'a Self> for GroupAffine<P> {
+    type Output = Self;
+
+    fn sub(self, other: &'a Self) -> Self {
+        let mut copy = self;
+        copy -= other;
+        copy
+    }
+}
+
+impl<'a, P: Parameters> SubAssign<&'a Self> for GroupAffine<P> {
+    fn sub_assign(&mut self, other: &'a Self) {
+        *self += &(-(*other));
+    }
+}
+
+impl<P: Parameters> Mul<P::ScalarField> for GroupAffine<P> {
+    type Output = Self;
+
+    fn mul(self, other: P::ScalarField) -> Self {
+        self.mul_bits(BitIteratorBE::new(other.into_repr())).into()
+    }
+}
+
+impl<P: Parameters> MulAssign<P::ScalarField> for GroupAffine<P> {
+    fn mul_assign(&mut self, other: P::ScalarField) {
+        *self = self.mul(other).into()
+    }
+}
+
 impl<P: Parameters> ToBytes for GroupAffine<P> {
     #[inline]
     fn write<W: Write>(&self, mut writer: W) -> IoResult<()> {
@@ -245,6 +302,20 @@ impl<P: Parameters> Default for GroupAffine<P> {
     }
 }
 
+impl<P: Parameters> Distribution<GroupAffine<P>> for Standard {
+    #[inline]
+    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> GroupAffine<P> {
+        loop {
+            let x = P::BaseField::rand(rng);
+            let greatest = rng.gen();
+
+            if let Some(p) = GroupAffine::from_x_coordinate(x, greatest) {
+                return p.scale_by_cofactor().into();
+            }
+        }
+    }
+}
+
 #[derive(Derivative)]
 #[derivative(
     Copy(bound = "P: Parameters"),
@@ -282,14 +353,14 @@ impl<P: Parameters> PartialEq for GroupProjective<P> {
         let z1 = self.z.square();
         let z2 = other.z.square();
 
-        !(self.x * &z2 != other.x * &z1 || self.y * &(z2 * &other.z) != other.y * &(z1 * &self.z))
+        !(self.x * z2 != other.x * z1 || self.y * (z2 * other.z) != other.y * (z1 * self.z))
     }
 }
 
 impl<P: Parameters> Distribution<GroupProjective<P>> for Standard {
     #[inline]
     fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> GroupProjective<P> {
-        let res = GroupProjective::prime_subgroup_generator() * &P::ScalarField::rand(rng);
+        let res = GroupProjective::prime_subgroup_generator() * P::ScalarField::rand(rng);
         debug_assert!(res.into_affine().is_in_correct_subgroup_assuming_on_curve());
         res
     }
@@ -351,7 +422,6 @@ impl<P: Parameters> Zero for GroupProjective<P> {
 impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
     type Affine = GroupAffine<P>;
     type BaseField = P::BaseField;
-    type ScalarField = P::ScalarField;
 
     #[inline]
     fn prime_subgroup_generator() -> Self {
@@ -400,8 +470,8 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
             )
         {
             // tmp := tmp * g.z; g.z := tmp * s = 1/z
-            let newtmp = tmp * &g.z;
-            g.z = tmp * &s;
+            let newtmp = tmp * g.z;
+            g.z = tmp * s;
             tmp = newtmp;
         }
         #[cfg(not(feature = "parallel"))]
@@ -410,7 +480,7 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
             for g in v.iter_mut().filter(|g| !g.is_normalized()) {
                 let z2 = g.z.square(); // 1/z
                 g.x *= &z2; // x/z^2
-                g.y *= &(z2 * &g.z); // y/z^3
+                g.y *= &(z2 * g.z); // y/z^3
                 g.z = P::BaseField::one(); // z = 1
             }
         }
@@ -422,81 +492,12 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
             v.par_iter_mut().filter(|g| !g.is_normalized()).for_each(|g| {
                 let z2 = g.z.square(); // 1/z
                 g.x *= &z2; // x/z^2
-                g.y *= &(z2 * &g.z); // y/z^3
+                g.y *= &(z2 * g.z); // y/z^3
                 g.z = P::BaseField::one(); // z = 1
             });
         }
     }
 
-    #[allow(clippy::many_single_char_names)]
-    fn double_in_place(&mut self) -> &mut Self {
-        if self.is_zero() {
-            return self;
-        }
-
-        if P::COEFF_A.is_zero() {
-            // A = X1^2
-            let mut a = self.x.square();
-
-            // B = Y1^2
-            let b = self.y.square();
-
-            // C = B^2
-            let mut c = b.square();
-
-            // D = 2*((X1+B)2-A-C)
-            let d = ((self.x + &b).square() - &a - &c).double();
-
-            // E = 3*A
-            let e = a + a.double_in_place();
-
-            // F = E^2
-            let f = e.square();
-
-            // Z3 = 2*Y1*Z1
-            self.z.mul_assign(&self.y);
-            self.z.double_in_place();
-
-            // X3 = F-2*D
-            self.x = f - &d - &d;
-
-            // Y3 = E*(D-X3)-8*C
-            self.y = (d - &self.x) * &e - c.double_in_place().double_in_place().double_in_place();
-            self
-        } else {
-            // http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian-0.html#doubling-dbl-2009-l
-            // XX = X1^2
-            let xx = self.x.square();
-
-            // YY = Y1^2
-            let yy = self.y.square();
-
-            // YYYY = YY^2
-            let mut yyyy = yy.square();
-
-            // ZZ = Z1^2
-            let zz = self.z.square();
-
-            // S = 2*((X1+YY)^2-XX-YYYY)
-            let s = ((self.x + &yy).square() - &xx - &yyyy).double();
-
-            // M = 3*XX+a*ZZ^2
-            let m = xx + &xx + &xx + &P::mul_by_a(&zz.square());
-
-            // T = M^2-2*S
-            let t = m.square() - &s.double();
-
-            // X3 = T
-            self.x = t;
-            // Y3 = M*(S-T)-8*YYYY
-            let old_y = self.y;
-            self.y = m * &(s - &t) - yyyy.double_in_place().double_in_place().double_in_place();
-            // Z3 = (Y1+Z1)^2-YY-ZZ
-            self.z = (old_y + &self.z).square() - &yy - &zz;
-            self
-        }
-    }
-
     #[allow(clippy::many_single_char_names)]
     fn add_assign_mixed(&mut self, other: &Self::Affine) {
         if other.is_zero() {
@@ -517,10 +518,10 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
         let z1z1 = self.z.square();
 
         // U2 = X2*Z1Z1
-        let u2 = other.x * &z1z1;
+        let u2 = other.x * z1z1;
 
         // S2 = Y2*Z1*Z1Z1
-        let s2 = (other.y * &self.z) * &z1z1;
+        let s2 = (other.y * self.z) * z1z1;
 
         if self.x == u2 && self.y == s2 {
             // The two points are equal, so we double.
@@ -529,23 +530,24 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
             // If we're adding -a and a together, self.z becomes zero as H becomes zero.
 
             // H = U2-X1
-            let h = u2 - &self.x;
+            let h = u2 - self.x;
 
             // HH = H^2
             let hh = h.square();
 
             // I = 4*HH
             let mut i = hh;
-            i.double_in_place().double_in_place();
+            i.double_in_place();
+            i.double_in_place();
 
             // J = H*I
-            let mut j = h * &i;
+            let mut j = h * i;
 
             // r = 2*(S2-Y1)
-            let r = (s2 - &self.y).double();
+            let r = (s2 - self.y).double();
 
             // V = X1*I
-            let v = self.x * &i;
+            let v = self.x * i;
 
             // X3 = r^2 - J - 2*V
             self.x = r.square();
@@ -556,7 +558,7 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
             // Y3 = r*(V-X3)-2*Y1*J
             j *= &self.y; // J = 2*Y1*J
             j.double_in_place();
-            self.y = v - &self.x;
+            self.y = v - self.x;
             self.y *= &r;
             self.y -= &j;
 
@@ -568,26 +570,6 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
         }
     }
 
-    fn mul_assign<S: Into<<Self::ScalarField as PrimeField>::BigInteger>>(&mut self, other: S) {
-        let mut res = Self::zero();
-
-        let mut found_one = false;
-
-        for i in BitIteratorBE::new(other.into()) {
-            if found_one {
-                res.double_in_place();
-            } else {
-                found_one = i;
-            }
-
-            if i {
-                res.add_assign(self);
-            }
-        }
-
-        *self = res;
-    }
-
     #[inline]
     fn into_affine(&self) -> GroupAffine<P> {
         (*self).into()
@@ -604,6 +586,93 @@ impl<P: Parameters> ProjectiveCurve for GroupProjective<P> {
     }
 }
 
+impl<P: Parameters> Group for GroupProjective<P> {
+    type ScalarField = P::ScalarField;
+
+    #[inline]
+    #[must_use]
+    fn double(&self) -> Self {
+        let mut tmp = *self;
+        tmp.double_in_place();
+        tmp
+    }
+
+    #[inline]
+    fn double_in_place(&mut self) {
+        if self.is_zero() {
+            return;
+        }
+
+        if P::COEFF_A.is_zero() {
+            // A = X1^2
+            let mut a = self.x.square();
+
+            // B = Y1^2
+            let b = self.y.square();
+
+            // C = B^2
+            let mut c = b.square();
+
+            // D = 2*((X1+B)2-A-C)
+            let d = ((self.x + b).square() - a - c).double();
+
+            // E = 3*A
+            let old_a = a;
+            a.double_in_place();
+            let e = old_a + a;
+
+            // F = E^2
+            let f = e.square();
+
+            // Z3 = 2*Y1*Z1
+            self.z *= &self.y;
+            self.z.double_in_place();
+
+            // X3 = F-2*D
+            self.x = f - d - d;
+
+            // Y3 = E*(D-X3)-8*C
+            c.double_in_place();
+            c.double_in_place();
+            c.double_in_place();
+            self.y = (d - self.x) * e - c;
+        } else {
+            // http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian-0.html#doubling-dbl-2009-l
+            // XX = X1^2
+            let xx = self.x.square();
+
+            // YY = Y1^2
+            let yy = self.y.square();
+
+            // YYYY = YY^2
+            let mut yyyy = yy.square();
+
+            // ZZ = Z1^2
+            let zz = self.z.square();
+
+            // S = 2*((X1+YY)^2-XX-YYYY)
+            let s = ((self.x + yy).square() - xx - yyyy).double();
+
+            // M = 3*XX+a*ZZ^2
+            let m = xx + xx + xx + P::mul_by_a(&zz.square());
+
+            // T = M^2-2*S
+            let t = m.square() - s.double();
+
+            // X3 = T
+            self.x = t;
+            // Y3 = M*(S-T)-8*YYYY
+            let old_y = self.y;
+            yyyy.double_in_place();
+            yyyy.double_in_place();
+            yyyy.double_in_place();
+            self.y = m * (s - t) - yyyy;
+            // Z3 = (Y1+Z1)^2-YY-ZZ
+            self.z = (old_y + self.z).square() - yy - zz;
+        }
+    }
+}
+
 impl<P: Parameters> Neg for GroupProjective<P> {
     type Output = Self;
 
@@ -617,6 +686,8 @@ impl<P: Parameters> Neg for GroupProjective<P> {
     }
 }
 
+impl_additive_ops_from_ref!(GroupProjective, Parameters);
+
 impl<'a, P: Parameters> Add<&'a Self> for GroupProjective<P> {
     type Output = Self;
 
@@ -651,16 +722,16 @@ impl<'a, P: Parameters> AddAssign<&'a Self> for GroupProjective<P> {
         let z2z2 = other.z.square();
 
         // U1 = X1*Z2Z2
-        let u1 = self.x * &z2z2;
+        let u1 = self.x * z2z2;
 
         // U2 = X2*Z1Z1
-        let u2 = other.x * &z1z1;
+        let u2 = other.x * z1z1;
 
         // S1 = Y1*Z2*Z2Z2
-        let s1 = self.y * &other.z * &z2z2;
+        let s1 = self.y * other.z * z2z2;
 
         // S2 = Y2*Z1*Z1Z1
-        let s2 = other.y * &self.z * &z1z1;
+        let s2 = other.y * self.z * z1z1;
 
         if u1 == u2 && s1 == s2 {
             // The two points are equal, so we double.
@@ -669,28 +740,28 @@ impl<'a, P: Parameters> AddAssign<&'a Self> for GroupProjective<P> {
             // If we're adding -a and a together, self.z becomes zero as H becomes zero.
 
             // H = U2-U1
-            let h = u2 - &u1;
+            let h = u2 - u1;
 
             // I = (2*H)^2
             let i = (h.double()).square();
 
             // J = H*I
-            let j = h * &i;
+            let j = h * i;
 
             // r = 2*(S2-S1)
-            let r = (s2 - &s1).double();
+            let r = (s2 - s1).double();
 
             // V = U1*I
-            let v = u1 * &i;
+            let v = u1 * i;
 
             // X3 = r^2 - J - 2*V
-            self.x = r.square() - &j - &(v.double());
+            self.x = r.square() - j - (v.double());
 
             // Y3 = r*(V - X3) - 2*S1*J
-            self.y = r * &(v - &self.x) - (s1 * &j).double_in_place();
+            self.y = r * (v - self.x) - (s1 * j).double();
 
             // Z3 = ((Z1+Z2)^2 - Z1Z1 - Z2Z2)*H
-            self.z = ((self.z + &other.z).square() - &z1z1 - &z2z2) * &h;
+            self.z = ((self.z + other.z).square() - z1z1 - z2z2) * h;
         }
     }
 }
@@ -712,21 +783,37 @@ impl<'a, P: Parameters> SubAssign<&'a Self> for GroupProjective<P> {
     }
 }
 
-impl<'a, P: Parameters> Mul<&'a P::ScalarField> for GroupProjective<P> {
+impl<P: Parameters> Mul<P::ScalarField> for GroupProjective<P> {
     type Output = Self;
 
+    /// Performs scalar multiplication of this element.
+    #[allow(clippy::suspicious_arithmetic_impl)]
     #[inline]
-    fn mul(self, other: &'a P::ScalarField) -> Self {
-        let mut copy = self;
-        copy *= other;
-        copy
+    fn mul(self, other: P::ScalarField) -> Self {
+        let mut res = Self::zero();
+
+        let mut found_one = false;
+
+        for i in BitIteratorBE::new(other.into_repr()) {
+            if found_one {
+                res.double_in_place();
+            } else {
+                found_one = i;
+            }
+
+            if i {
+                res += self;
+            }
+        }
+
+        res
     }
 }
 
-impl<'a, P: Parameters> MulAssign<&'a P::ScalarField> for GroupProjective<P> {
-    #[inline]
-    fn mul_assign(&mut self, other: &'a P::ScalarField) {
-        <GroupProjective<P> as ProjectiveCurve>::mul_assign(self, *other);
+impl<P: Parameters> MulAssign<P::ScalarField> for GroupProjective<P> {
+    /// Performs scalar multiplication of this element.
+    fn mul_assign(&mut self, other: P::ScalarField) {
+        *self = *self * other
     }
 }
 
@@ -759,10 +846,10 @@ impl<P: Parameters> From<GroupProjective<P>> for GroupAffine<P> {
             let zinv_squared = zinv.square();
 
             // X/Z^2
-            let x = p.x * &zinv_squared;
+            let x = p.x * zinv_squared;
 
             // Y/Z^3
-            let y = p.y * &(zinv_squared * &zinv);
+            let y = p.y * (zinv_squared * zinv);
 
             GroupAffine::new(x, y, false)
         }
